import { uint32 } from "./number";
import { Anchor } from "./anchor";
import { Position } from "./position";
import { Concatenable, ConcatenableLength } from "./concatenable";
import { IntInterval } from "./intinterval";
/**
 * Possible relation between two blocks.
 *
 * Same semantic as {@link IntervalOrdering }.
 * @example
 * [0, 2] |⊂ 1[_, _]
 */
export declare const enum BlockOrdering {
    BEFORE = -5,
    PREPENDABLE = -4,
    OVERLAPPING_BEFORE = -3,
    SPLITTED_BY = -2,
    INCLUDED_BY = -1,
    EQUAL = 0,
    INCLUDING = 1,
    SPLITTING = 2,
    OVERLAPPING_AFTER = 3,
    APPENDABLE = 4,
    AFTER = 5,
}
/**
 * Represent the removal of a segment.
 */
export declare type LengthBlock<P extends Position<P>> = Block<P, ConcatenableLength>;
/**
 * Segment of concatenable elements generated by a same participant.
 *
 * A segment can be represented as a set of int successive positions
 * (see {@link Position }) and an attached content (items).
 */
export declare class Block<P extends Position<P>, E extends Concatenable<E>> {
    /**
     * @param lowerPosition {@link lowerPosition }
     * @param items {@link items }
     */
    constructor(lowerPosition: P, items: E);
    /**
     * First position of the block.
     */
    readonly lowerPosition: P;
    /**
     * Content of the block.
     */
    readonly items: E;
    /**
     * Length of {@link Block#items }
     */
    readonly length: uint32;
    /**
     * @example
     * b.nthPosition(0) == b.lowerPositions
     * b.nthPosition(b.length - 1) == b.upperPosition
     *
     * @param nth 0-based index
     * @return {@link nth } -th position of the block.
     */
    nthPosition(nth: uint32): P;
    /**
     * Last position of the block.
     */
    readonly upperPosition: P;
    /**
     * @param index 0-based index
     *      {@link index } < {@link Block#length }
     * @return Anchor at {@link index }
     */
    anchorAt(index: uint32): Anchor<P>;
    /**
     * Globally unique identifier of the author which generated this block.
     */
    readonly replica: uint32;
    /**
     * When each position of this block were generated.
     */
    readonly seqs: IntInterval;
    /**
     * @param other splitting block
     *      {@link other } must split this
     * @return 0-based index where {@link other } splits this block.
     */
    splittingIndex(other: Block<P, E> | LengthBlock<P>): uint32;
    /**
     * @return Length block representation of this block.
     */
    toLengthBlock(): LengthBlock<P>;
    /**
     * Hash code.
     * Note that the content is not take into account.
     */
    readonly structuralDigest: uint32;
    /**
     * @return Is this block a length block?
     */
    isLengthBlock(): this is LengthBlock<P>;
    /**
     * @param other
     * @return Are this and {@link other } an intersection?
     */
    hasIntersection(other: Block<P, E> | LengthBlock<P>): boolean;
    /**
     * @example
     * a.compare(b) == Ordering.Before if a < b
     *
     * @param other
     * @return this [Order relation] {@link other}.
     */
    compare(other: Block<P, E> | LengthBlock<P>): BlockOrdering;
    /**
     * @example
     * Block(p, "ab").append(Block(_, "cd")) == Block(p, "abcd")
     *
     * @param other
     *      {@link other } is appendable to this.
     * @return other appended to this.
     */
    append(other: Block<P, E>): Block<P, E>;
    /**
     * @example
     * Block(p, "ab").leftSplitAt(1) == Block(p, "a")
     * Block(p, "ab").leftSplitAt(2) == Block(p, "ab")
     *
     * @param index 0-based index
     *      0 < {@link index } <= {@link Block#length }
     * @return Left split.
     */
    leftSplitAt(index: uint32): Block<P, E>;
    /**
     * @example
     * Block(p, "ab").rightSplitAt(1) == Block(p+1, "b")
     * Block(p, "ab").rightSplitAt(0) == Block(p, "ab")
     *
     * @param index 0-based index
     *      {@link index } < this.length
     * @return Right split.
     */
    rightSplitAt(index: uint32): Block<P, E>;
    /**
     * @example
     * b.splitAt(n) == [b.leftSplitAt(n), b.rightSplitAt(n)]
     *
     * @param index 0-based index
     *      0 < {@link index } < {@link Block#length }
     * @return Left and right splits.
     */
    splitAt(index: uint32): [Block<P, E>, Block<P, E>];
    /**
     * @param other splitting block
     *      {@link other } must split this
     * @return Left and right splits.
     */
    splitWith(other: Block<P, E> | LengthBlock<P>): [Block<P, E>, Block<P, E>];
    /**
     * @param other
     *      this includes or overlaps after {@link other }
     * @return Part of this block which can be append to {@link block }.
     */
    appendable(other: Block<P, E> | LengthBlock<P>): Block<P, E>;
    /**
     * @param other
     *      this includes or overlaps before {@link other }
     * @return Part of this block which can be prepend to {@link block }.
     */
    prependable(other: Block<P, E> | LengthBlock<P>): Block<P, E>;
    /**
     * @param other
     *      this and {@link other } intersect.
     * @return Intersection part between this and {@link other }.
     */
    intersection(other: Block<P, E> | LengthBlock<P>): Block<P, E>;
    /**
     * @example
     * Block(p, "abc").remove(Block(p, "a")) == [undefined, Block(p+1, "bc")]
     * Block(p, "abc").remove(Block(p+1, "b")) == [Block(_, "a"), Block(p+2, "c")]
     * Block(p, "abc").remove(Block(p+2, "c")) == [Block(p, "ab"), undefined]
     * Block(p, "abc").remove(Block(p+3, "d")) == [undefined, undefined]
     *
     * @param other
     * @return Left and right remaining after removing from this the
     *  intersection between this and {@link other }.
     */
    remove(other: Block<P, E> | LengthBlock<P>): [Block<P, E> | undefined, Block<P, E> | undefined];
}
