/*
    Copyright (C) 2018  Victorien Elvinger

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.
*/
var __read = (this && this.__read) || function (o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
};
import { assert, heavyAssert, readonly } from "./assert";
import { digestOf, isUint32 } from "./number";
import { Anchor } from "./anchor";
import { ConcatenableLength } from "./concatenable";
import { IntInterval } from "./intinterval";
/**
 * Map IntervalOrdering to BlockOrdering.
 */
var intervalOrderingAsBlockOrdering = readonly((_a = {},
    _a[-4 /* BEFORE */] = -5 /* BEFORE */,
    _a[-3 /* PREPENDABLE */] = -4 /* PREPENDABLE */,
    _a[-2 /* OVERLAPPING_BEFORE */] = -3 /* OVERLAPPING_BEFORE */,
    _a[1 /* INCLUDING */] = 1 /* INCLUDING */,
    _a[0 /* EQUAL */] = 0 /* EQUAL */,
    _a[-1 /* INCLUDED_BY */] = -1 /* INCLUDED_BY */,
    _a[2 /* OVERLAPPING_AFTER */] = 3 /* OVERLAPPING_AFTER */,
    _a[3 /* APPENDABLE */] = 4 /* APPENDABLE */,
    _a[4 /* AFTER */] = 5 /* AFTER */,
    _a));
/**
 * Segment of concatenable elements generated by a same participant.
 *
 * A segment can be represented as a set of int successive positions
 * (see {@link Position }) and an attached content (items).
 */
var Block = /** @class */ (function () {
    /**
     * @param lowerPosition {@link lowerPosition }
     * @param items {@link items }
     */
    function Block(lowerPosition, items) {
        assert(function () { return items.length > 0; }, "items.length > 0");
        assert(function () { return lowerPosition.hasIntSuccessor(items.length - 1); }, "all positions are valid");
        this.lowerPosition = lowerPosition;
        this.items = items;
    }
    Object.defineProperty(Block.prototype, "length", {
        /**
         * Length of {@link Block#items }
         */
        get: function () {
            return this.items.length;
        },
        enumerable: true,
        configurable: true
    });
    /**
     * @example
     * b.nthPosition(0) == b.lowerPositions
     * b.nthPosition(b.length - 1) == b.upperPosition
     *
     * @param nth 0-based index
     * @return {@link nth } -th position of the block.
     */
    Block.prototype.nthPosition = function (nth) {
        var _this = this;
        assert(function () { return isUint32(nth); }, "nth ∈ uint32");
        assert(function () { return nth < _this.length; }, "nth < this.length");
        if (nth === 0) {
            return this.lowerPosition;
        }
        else {
            return this.lowerPosition.intSuccessor(nth);
        }
    };
    Object.defineProperty(Block.prototype, "upperPosition", {
        /**
         * Last position of the block.
         */
        get: function () {
            return this.nthPosition(this.length - 1);
        },
        enumerable: true,
        configurable: true
    });
    /**
     * @param index 0-based index
     *      {@link index } < {@link Block#length }
     * @return Anchor at {@link index }
     */
    Block.prototype.anchorAt = function (index) {
        var _this = this;
        assert(function () { return isUint32(index); }, "index ∈ uint32");
        assert(function () { return index < _this.length; }, "index < this.length");
        return new Anchor(this.nthPosition(index));
    };
    Object.defineProperty(Block.prototype, "replica", {
        /**
         * Globally unique identifier of the author which generated this block.
         */
        get: function () {
            return this.lowerPosition.replica;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Block.prototype, "seqs", {
        /**
         * When each position of this block were generated.
         */
        get: function () {
            return IntInterval.fromLength(this.lowerPosition.seq, this.length);
        },
        enumerable: true,
        configurable: true
    });
    /**
     * @param other splitting block
     *      {@link other } must split this
     * @return 0-based index where {@link other } splits this block.
     */
    Block.prototype.splittingIndex = function (other) {
        var _this = this;
        heavyAssert(function () { return _this.compare(other) === -2 /* SPLITTED_BY */; }, "other splits this");
        var _a = __read(this.lowerPosition.intDistance(other.lowerPosition), 1), dist = _a[0];
        return dist + 1;
    };
    /**
     * @return Length block representation of this block.
     */
    Block.prototype.toLengthBlock = function () {
        var concatLength = new ConcatenableLength(this.length);
        return new Block(this.lowerPosition, concatLength);
    };
    Object.defineProperty(Block.prototype, "structuralDigest", {
        /**
         * Hash code.
         * Note that the content is not take into account.
         */
        get: function () {
            return digestOf([this.lowerPosition.digest, this.length]);
        },
        enumerable: true,
        configurable: true
    });
    // Status
    /**
     * @return Is this block a length block?
     */
    Block.prototype.isLengthBlock = function () {
        return this.items instanceof ConcatenableLength;
    };
    /**
     * @param other
     * @return Are this and {@link other } an intersection?
     */
    Block.prototype.hasIntersection = function (other) {
        var cmp = this.compare(other);
        return cmp === -3 /* OVERLAPPING_BEFORE */ ||
            cmp === 3 /* OVERLAPPING_AFTER */ || cmp === 0 /* EQUAL */ ||
            cmp === 1 /* INCLUDING */ || cmp === -1 /* INCLUDED_BY */;
    };
    /**
     * @example
     * a.compare(b) == Ordering.Before if a < b
     *
     * @param other
     * @return this [Order relation] {@link other}.
     */
    Block.prototype.compare = function (other) {
        var baseCmp = this.lowerPosition.compareBase(other.lowerPosition);
        if (baseCmp === -2 /* BEFORE */) {
            return -5 /* BEFORE */; // Before meta block
        }
        else if (baseCmp === 2 /* AFTER */) {
            return 5 /* AFTER */; // After meta block
        }
        else {
            var _a = __read(this.lowerPosition.intDistance(other.lowerPosition), 2), dist = _a[0], order = _a[1];
            if (baseCmp === 0 /* EQUAL */) {
                var lower1 = (order === -1 /* BEFORE */) ? 0 : dist;
                var thisInterval = IntInterval.fromLength(lower1, this.length);
                var lower2 = (order === -1 /* BEFORE */) ? dist : 0;
                var otherInterval = IntInterval.fromLength(lower2, other.length);
                var intervalCmp = thisInterval.compare(otherInterval);
                return intervalOrderingAsBlockOrdering[intervalCmp];
            }
            else if (baseCmp === -1 /* PREFIXING */) {
                if (order === 1 /* AFTER */) {
                    return 5 /* AFTER */;
                }
                else if (dist >= (this.length - 1)) {
                    return -5 /* BEFORE */;
                }
                else {
                    return -2 /* SPLITTED_BY */;
                }
            }
            else {
                if (order === -1 /* BEFORE */) {
                    return -5 /* BEFORE */;
                }
                else if (dist >= (other.length - 1)) {
                    return 5 /* AFTER */;
                }
                else {
                    return 2 /* SPLITTING */;
                }
            }
        }
    };
    // Derivation
    /**
     * @example
     * Block(p, "ab").append(Block(_, "cd")) == Block(p, "abcd")
     *
     * @param other
     *      {@link other } is appendable to this.
     * @return other appended to this.
     */
    Block.prototype.append = function (other) {
        var _this = this;
        heavyAssert(function () { return other.compare(_this) === 4 /* APPENDABLE */; }, "other is appendable to this");
        return new Block(this.lowerPosition, this.items.concat(other.items));
    };
    /**
     * @example
     * Block(p, "ab").leftSplitAt(1) == Block(p, "a")
     * Block(p, "ab").leftSplitAt(2) == Block(p, "ab")
     *
     * @param index 0-based index
     *      0 < {@link index } <= {@link Block#length }
     * @return Left split.
     */
    Block.prototype.leftSplitAt = function (index) {
        var _this = this;
        assert(function () { return isUint32(index); }, "index ∈ uint32");
        assert(function () { return 0 < index && index <= _this.length; }, "0 < index <= this.length");
        var leftItems = this.items.slice(0, index);
        return new Block(this.lowerPosition, leftItems);
    };
    /**
     * @example
     * Block(p, "ab").rightSplitAt(1) == Block(p+1, "b")
     * Block(p, "ab").rightSplitAt(0) == Block(p, "ab")
     *
     * @param index 0-based index
     *      {@link index } < this.length
     * @return Right split.
     */
    Block.prototype.rightSplitAt = function (index) {
        var _this = this;
        assert(function () { return isUint32(index); }, "index ∈ uint32");
        assert(function () { return index < _this.length; }, "0 <= index < this.length");
        var rightItems = this.items.slice(index, this.length);
        var rightPosition = this.nthPosition(index);
        return new Block(rightPosition, rightItems);
    };
    /**
     * @example
     * b.splitAt(n) == [b.leftSplitAt(n), b.rightSplitAt(n)]
     *
     * @param index 0-based index
     *      0 < {@link index } < {@link Block#length }
     * @return Left and right splits.
     */
    Block.prototype.splitAt = function (index) {
        var _this = this;
        assert(function () { return isUint32(index); }, "index ∈ uint32");
        assert(function () { return 0 < index && index < _this.length; }, "0 < index < this.length");
        return [this.leftSplitAt(index), this.rightSplitAt(index)];
    };
    /**
     * @param other splitting block
     *      {@link other } must split this
     * @return Left and right splits.
     */
    Block.prototype.splitWith = function (other) {
        var _this = this;
        assert(function () { return _this.compare(other) === -2 /* SPLITTED_BY */; }, "other splits this");
        return this.splitAt(this.splittingIndex(other));
    };
    /**
     * @param other
     *      this includes or overlaps after {@link other }
     * @return Part of this block which can be append to {@link block }.
     */
    Block.prototype.appendable = function (other) {
        var _this = this;
        heavyAssert(function () { return (function (cmp) {
            return cmp === 3 /* OVERLAPPING_AFTER */ ||
                cmp === 1 /* INCLUDING */;
        })(_this.compare(other)); }, "this has an appendable segment to other");
        var _a = __read(this.lowerPosition.intDistance(other.lowerPosition), 2), dist = _a[0], order = _a[1];
        if (order === -1 /* BEFORE */) {
            return this.rightSplitAt(other.length + dist);
        }
        else {
            return this.rightSplitAt(other.length - dist);
        }
    };
    /**
     * @param other
     *      this includes or overlaps before {@link other }
     * @return Part of this block which can be prepend to {@link block }.
     */
    Block.prototype.prependable = function (other) {
        var _this = this;
        heavyAssert(function () { return (function (cmp) {
            return cmp === -3 /* OVERLAPPING_BEFORE */ ||
                cmp === 1 /* INCLUDING */;
        })(_this.compare(other)); }, "this has a prependable segment to other");
        var _a = __read(this.lowerPosition.intDistance(other.lowerPosition), 1), dist = _a[0];
        return this.leftSplitAt(dist);
    };
    /**
     * @param other
     *      this and {@link other } intersect.
     * @return Intersection part between this and {@link other }.
     */
    Block.prototype.intersection = function (other) {
        var _this = this;
        heavyAssert(function () { return _this.hasIntersection(other); }, "this and other intersect.");
        var _a = __read(this.lowerPosition.intDistance(other.lowerPosition), 2), dist = _a[0], order = _a[1];
        if (order === -1 /* BEFORE */) {
            var pos = other.lowerPosition;
            var l = Math.min(this.length - dist, other.length);
            var elements = this.items.slice(dist, dist + l);
            return new Block(pos, elements);
        }
        else {
            var pos = this.lowerPosition;
            var l = Math.min(other.length - dist, this.length);
            var elements = this.items.slice(0, l);
            return new Block(pos, elements);
        }
    };
    /**
     * @example
     * Block(p, "abc").remove(Block(p, "a")) == [undefined, Block(p+1, "bc")]
     * Block(p, "abc").remove(Block(p+1, "b")) == [Block(_, "a"), Block(p+2, "c")]
     * Block(p, "abc").remove(Block(p+2, "c")) == [Block(p, "ab"), undefined]
     * Block(p, "abc").remove(Block(p+3, "d")) == [undefined, undefined]
     *
     * @param other
     * @return Left and right remaining after removing from this the
     *  intersection between this and {@link other }.
     */
    Block.prototype.remove = function (other) {
        var _this = this;
        heavyAssert(function () { return _this.hasIntersection(other); }, "this and other intersect.");
        var _a = __read(this.lowerPosition.intDistance(other.lowerPosition), 2), dist = _a[0], order = _a[1];
        if (order === -1 /* BEFORE */) {
            var removedLength = this.length - dist;
            if (removedLength > other.length) {
                return [this.prependable(other), this.appendable(other)];
            }
            else {
                return [this.prependable(other), undefined];
            }
        }
        else {
            var removedLength = other.length - dist;
            if (removedLength >= this.length) {
                return [undefined, undefined];
            }
            else {
                return [undefined, this.appendable(other)];
            }
        }
    };
    return Block;
}());
export { Block };
var _a;
//# sourceMappingURL=block.js.map