/*
    Copyright (C) 2018  Victorien Elvinger

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.
*/
import { assert } from "./assert";
import { isUint32 } from "./number";
/**
 * Interval of uint32.
 */
var IntInterval = /** @class */ (function () {
    /**
     * @param lower {@link IntInterval#lower }
     * @param length  {@link IntInterval#length }
     */
    function IntInterval(lower, length) {
        this.lower = lower;
        this.length = length;
    }
    /**
     * @param lower {@link IntInterval#lower }
     * @param length {@link IntInterval#length }
     * @return New interval.
     */
    IntInterval.fromLength = function (lower, length) {
        assert(function () { return isUint32(lower); }, "lower ∈ uint32");
        assert(function () { return isUint32(length); }, "length ∈ uint32");
        assert(function () { return length > 0; }, "length > 0");
        //assert(() => isUint32(lower + length))
        return new IntInterval(lower, length);
    };
    /**
     * @param lower {@link IntInterval#lower }
     * @param upper {@link IntInterval#length }
     * @return New interval.
     */
    IntInterval.fromBounds = function (lower, upper) {
        assert(function () { return isUint32(lower); }, "lower ∈ uint32");
        assert(function () { return isUint32(upper); }, "upper ∈ uint32");
        assert(function () { return lower <= upper; }, "lower <= upper");
        return IntInterval.fromLength(lower, upper - lower + 1);
    };
    /**
     * @example
     * nth(0) == lower
     * nth(length - 1) == upper
     *
     * @param nth 0-based
     * @return n-th element.
     */
    IntInterval.prototype.nth = function (nth) {
        var _this = this;
        assert(function () { return isUint32(nth); }, "nth ∈ uint32");
        assert(function () { return nth < _this.length; }, "valid nth");
        assert(function () { return isUint32(_this.lower + nth); }, "No integer overflow");
        return this.lower + nth;
    };
    Object.defineProperty(IntInterval.prototype, "upper", {
        /**
         * Upper bound.
         */
        get: function () {
            return this.nth(this.length - 1);
        },
        enumerable: true,
        configurable: true
    });
    // Derivation
    /**
     * @param other
     * @return {@link other } appended to this.
     */
    IntInterval.prototype.append = function (other) {
        var _this = this;
        assert(function () { return other.compare(_this) === 3 /* APPENDABLE */; }, "other is appendable to this");
        return new IntInterval(this.lower, this.length + other.length);
    };
    /**
     * @param index where split.
     *      0 < index < length
     * @return Right and left splits.
     */
    IntInterval.prototype.splitAt = function (index) {
        var _this = this;
        assert(function () { return isUint32(index); }, "index ∈ uint32");
        assert(function () { return 0 < index && index < _this.length; }, "0 < index < this.length");
        var leftInterval = new IntInterval(this.lower, index);
        var rightInterval = new IntInterval(this.lower + index, this.length - index);
        return [leftInterval, rightInterval];
    };
    // Status
    /**
     * @param other
     * @return this [order relation] other.
     */
    IntInterval.prototype.compare = function (other) {
        if (this.upper < other.lower) {
            if (this.upper + 1 === other.lower) {
                return -3 /* PREPENDABLE */;
            }
            else {
                return -4 /* BEFORE */;
            }
        }
        else if (other.upper < this.lower) {
            if (other.upper + 1 === this.lower) {
                return 3 /* APPENDABLE */;
            }
            else {
                return 4 /* AFTER */;
            }
        }
        else {
            if (this.lower === other.lower) {
                if (this.upper === other.upper) {
                    return 0 /* EQUAL */;
                }
                else if (this.upper < other.upper) {
                    return -1 /* INCLUDED_BY */;
                }
                else {
                    return 1 /* INCLUDING */;
                }
            }
            else if (this.lower < other.lower) {
                if (other.upper <= this.upper) {
                    return 1 /* INCLUDING */;
                }
                else {
                    return -2 /* OVERLAPPING_BEFORE */;
                }
            }
            else {
                if (this.upper <= other.upper) {
                    return -1 /* INCLUDED_BY */;
                }
                else {
                    return 2 /* OVERLAPPING_AFTER */;
                }
            }
        }
    };
    return IntInterval;
}());
export { IntInterval };
//# sourceMappingURL=intinterval.js.map