import { Block, LengthBlock } from "../core/block";
import { BlockFactory } from "../core/blockfactory";
import { Concatenable } from "../core/concatenable";
import { Position } from "../core/position";
import { Insertion, Deletion } from "../core/localoperation";
import { uint32 } from "../core/number";
export declare abstract class Linkable<P extends Position<P>, E extends Concatenable<E>> {
    /**
     * @param right {@link Linkable#right }
     */
    constructor(right?: Cell<P, E>);
    /**
     * Next cell.
     */
    right?: Cell<P, E>;
    reduceBlock<U>(f: (acc: U, b: Block<P, E>) => U, prefix: U): U;
    /**
     * [Mutation]
     * Insert {@link rblock } to right and preserve the chain.
     *
     * @param rblock block to insert.
     * @return inserted cell with {@link rblock } as content.
     */
    protected insertRight(rblock: Block<P, E>): Cell<P, E>;
    /**
     * [Mutation]
     * Insert the parts of {@link iBlock } which are not already inserted.
     *
     * @param iBlock block to insert.
     * @param index 0-based index
     * @return Performed modifications in terms of local insertions.
     *  The n+1 -th insertion depends on the effect of the n -th insertion.
     */
    insert(iBlock: Block<P, E>, index: uint32): Insertion<E>[];
    /**
     * [Mutation]
     * Insert {@link items } at {@link index}.
     *
     * @param index 0-based index; Where to insert.
     * @param items elements to insert.
     * @param factory strategy of block generation.
     * @return Delta which represents the insertion.
     */
    insertAt(index: uint32, items: E, factory: BlockFactory<P>): [Block<P, E>, BlockFactory<P>];
    /**
     * [Mutation]
     * Remove a number of {@link length } elements from {@link index }
     *
     * @param index 0-based index.
     * @param length Number of elements to remove.
     * @return Delta which represents the deletion.
     */
    removeAt(index: uint32, length: uint32): LengthBlock<P>[];
    /**
     * [Mutation]
     * Remove {@link dBlock }.
     * @param dBlock block to remove.
     * @param index index of the current block in the chain.
     * @return Performed modifications in terms of local deletions.
     *  The n+1 -th deletion depends on the n -th deletion.
     */
    remove(dBlock: LengthBlock<P> | Block<P, E>, index: uint32): Deletion[];
}
export declare class Sentinel<P extends Position<P>, E extends Concatenable<E>> extends Linkable<P, E> {
}
export declare class Cell<P extends Position<P>, E extends Concatenable<E>> extends Linkable<P, E> {
    /**
     * @param block {@link Cell#block }
     * @param right {@link Linkable#right }
     */
    constructor(block: Block<P, E>, right?: Cell<P, E>);
    /**
     * Block attached to this cell.
     */
    readonly block: Block<P, E>;
}
