import { uint32 } from "../core/number";
import { Position, BaseOrdering } from "../core/position";
import { SimplePositionPart } from "./simplepositionpart";
import { Ordering } from "../core/ordering";
/**
 * A position can be represented as a lexicographic list of
 * lexicographic triplets (priority, replica, seq).
 * Theses triplets are the parts of the position. The depth of a triplet
 * correponds to its 0-based index in the list.
 *
 * The dot (replica, seq) uniquely identifies the position.
 *
 * The set of positions is dense, hence you can always author a new position
 * between two distinct positions. A position can be generated between two
 * int-successive positions (see {@link Position}) by suffixing a new triplet
 * to the lower position.
 */
export declare class SimplePosition implements Position<SimplePosition> {
    /**
     * @param parts {@link SimplePosition#parts }
     */
    protected constructor(parts: ReadonlyArray<SimplePositionPart>);
    /**
     * @param parts {@link SimplePosition#parts }
     *  The last part must be distinct of SimplePositionPart.BOTTOM and
     *  SimplePositionPart.TOP.
     * @return Position with {@link parts } as {@link SimplePosition#parts }.
     */
    static from(parts: ReadonlyArray<SimplePositionPart>): SimplePosition;
    /**
     * Lowest position.
     */
    static readonly BOTTOM: SimplePosition;
    /**
     * Greatest position.
     */
    static readonly TOP: SimplePosition;
    /**
     * @param seq The last seq of the new position
     * @return Position with the same base as this,
     *  but with a different seq
     */
    withSeq(seq: uint32): SimplePosition;
    /** @override */
    intSuccessor(n: uint32): SimplePosition;
    /**
     * Parts of this position.
     */
    readonly parts: ReadonlyArray<SimplePositionPart>;
    /**
     * Last part of {@link SimplePositionPart#parts }.
     */
    readonly lastPart: SimplePositionPart;
    /**
     * Number of part in {@link SimplePositionPart#parts }.
     */
    readonly depth: uint32;
    /** @override */
    readonly replica: uint32;
    /** @override */
    readonly seq: uint32;
    /** @override */
    readonly blockIdentifier: ReadonlyArray<uint32>;
    /** @override */
    intDistance(other: SimplePosition): [uint32, Ordering];
    /**
     * Hash code.
     */
    readonly digest: uint32;
    /** @override */
    hasIntSuccessor(n: uint32): boolean;
    /** @override */
    compareBase(other: SimplePosition): BaseOrdering;
    /** @override */
    isBaseEqual(other: SimplePosition): boolean;
    /** @override */
    compare(other: SimplePosition): Ordering;
    /** @override */
    isEqual(other: SimplePosition): boolean;
}
