/*
    Copyright (C) 2018  Victorien Elvinger

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.
*/
var __read = (this && this.__read) || function (o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
};
var __spread = (this && this.__spread) || function () {
    for (var ar = [], i = 0; i < arguments.length; i++) ar = ar.concat(__read(arguments[i]));
    return ar;
};
import { assert, heavyAssert } from "../core/assert";
import { absoluteSubstraction, compareUint32, digestOf, isUint32, UINT32_BOTTOM, UINT32_TOP, } from "../core/number";
import { baseOrderingInversion } from "../core/position";
import { SimplePositionPart } from "./simplepositionpart";
import { orderingInversion } from "../core/ordering";
/**
 * A position can be represented as a lexicographic list of
 * lexicographic triplets (priority, replica, seq).
 * Theses triplets are the parts of the position. The depth of a triplet
 * correponds to its 0-based index in the list.
 *
 * The dot (replica, seq) uniquely identifies the position.
 *
 * The set of positions is dense, hence you can always author a new position
 * between two distinct positions. A position can be generated between two
 * int-successive positions (see {@link Position}) by suffixing a new triplet
 * to the lower position.
 */
var SimplePosition = /** @class */ (function () {
    /**
     * @param parts {@link SimplePosition#parts }
     */
    function SimplePosition(parts) {
        assert(function () { return parts.length > 0; }, "parts must not be empty");
        this.parts = parts;
    }
    /**
     * @param parts {@link SimplePosition#parts }
     *  The last part must be distinct of SimplePositionPart.BOTTOM and
     *  SimplePositionPart.TOP.
     * @return Position with {@link parts } as {@link SimplePosition#parts }.
     */
    SimplePosition.from = function (parts) {
        var lastPart = parts[parts.length - 1];
        assert(function () { return lastPart.priority !== UINT32_BOTTOM && lastPart.priority !== UINT32_TOP; }, "priority ∉ {UINT32_BOTTOM, UINT32_TOP}");
        assert(function () { return lastPart.replica !== UINT32_TOP; }, "replica != UINT32_TOP. This is reserved for BOTTOM and TOP.");
        return new SimplePosition(parts);
    };
    // Derivation
    /**
     * @param seq The last seq of the new position
     * @return Position with the same base as this,
     *  but with a different seq
     */
    SimplePosition.prototype.withSeq = function (seq) {
        assert(function () { return isUint32(seq); }, "offset ∈ uint32");
        var parts = __spread(this.parts);
        var lastIndex = parts.length - 1;
        parts[lastIndex] = parts[lastIndex].withSeq(seq);
        return new SimplePosition(parts);
    };
    /** @override */
    SimplePosition.prototype.intSuccessor = function (n) {
        var _this = this;
        assert(function () { return isUint32(n); }, "n ∈ uint32");
        assert(function () { return _this.hasIntSuccessor(n); }, "this has a n-th successor");
        return this.withSeq(this.seq + n);
    };
    Object.defineProperty(SimplePosition.prototype, "lastPart", {
        /**
         * Last part of {@link SimplePositionPart#parts }.
         */
        get: function () {
            return this.parts[this.parts.length - 1];
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(SimplePosition.prototype, "depth", {
        /**
         * Number of part in {@link SimplePositionPart#parts }.
         */
        get: function () {
            return this.parts.length;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(SimplePosition.prototype, "replica", {
        /** @override */
        get: function () {
            return this.lastPart.replica;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(SimplePosition.prototype, "seq", {
        /** @override */
        get: function () {
            return this.lastPart.seq;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(SimplePosition.prototype, "blockIdentifier", {
        /** @override */
        get: function () {
            // TODO use a typed array and then a typeable array as return type?
            var result = this.parts.reduce(function (acc, part) { return (acc.concat(part.asTuple())); }, []);
            result.pop(); // remove last seq
            return result;
        },
        enumerable: true,
        configurable: true
    });
    /** @override */
    SimplePosition.prototype.intDistance = function (other) {
        var _this = this;
        if (this.depth > other.depth) {
            var _a = __read(other.intDistance(this), 2), dist = _a[0], order = _a[1];
            return [dist, orderingInversion[order]];
        }
        else {
            heavyAssert(function () { return (function (cmp) { return cmp === -1 /* PREFIXING */ ||
                cmp === 0 /* EQUAL */; })(_this.compareBase(other)); }, "MARK");
            var otherSeq = other.parts[this.depth - 1].seq;
            return [
                absoluteSubstraction(this.seq, otherSeq),
                compareUint32(this.seq, otherSeq)
            ];
        }
    };
    Object.defineProperty(SimplePosition.prototype, "digest", {
        /**
         * Hash code.
         */
        get: function () {
            return digestOf(this.parts.map(function (part) { return part.digest; }));
        },
        enumerable: true,
        configurable: true
    });
    // Status
    /** @override */
    SimplePosition.prototype.hasIntSuccessor = function (n) {
        assert(function () { return isUint32(n); }, "n ∈ uint32");
        return isUint32(this.seq + n);
    };
    /** @override */
    SimplePosition.prototype.compareBase = function (other) {
        if (this.depth > other.depth) {
            return baseOrderingInversion[other.compareBase(this)];
        }
        else if (this.replica === other.replica && this.seq === other.seq) {
            return 0 /* EQUAL */;
        }
        else {
            var i = 0;
            var baseCmp = 0 /* EQUAL */;
            while (i < (this.depth - 1) && baseCmp === 0 /* EQUAL */) {
                baseCmp = this.parts[i].compare(other.parts[i]);
                i++;
            }
            if (baseCmp === 0 /* EQUAL */) {
                baseCmp = this.parts[i].compareBase(other.parts[i]);
            }
            switch (baseCmp) {
                case -1 /* BEFORE */:
                    return -2 /* BEFORE */;
                case 0 /* EQUAL */:
                    if (this.depth === other.depth) {
                        return 0 /* EQUAL */;
                    }
                    else {
                        return -1 /* PREFIXING */;
                    }
                default:
                    return 2 /* AFTER */;
            }
        }
    };
    /** @override */
    SimplePosition.prototype.isBaseEqual = function (other) {
        return this.compareBase(other) === 0 /* EQUAL */;
    };
    /** @override */
    SimplePosition.prototype.compare = function (other) {
        if (this.depth > other.depth) {
            return orderingInversion[other.compare(this)];
        }
        else if (this.isEqual(other)) {
            return 0 /* EQUAL */;
        }
        else {
            var i = 0;
            var cmp = void 0;
            do {
                cmp = this.parts[i].compare(other.parts[i]);
                i++;
            } while (i < this.depth && cmp === 0 /* EQUAL */);
            if (cmp === 0 /* EQUAL */) {
                return compareUint32(this.depth, other.depth);
            }
            else {
                return cmp;
            }
        }
    };
    /** @override */
    SimplePosition.prototype.isEqual = function (other) {
        return this.replica === other.replica && this.seq === other.seq;
    };
    /**
     * Lowest position.
     */
    SimplePosition.BOTTOM = new SimplePosition([SimplePositionPart.BOTTOM]);
    /**
     * Greatest position.
     */
    SimplePosition.TOP = new SimplePosition([SimplePositionPart.TOP]);
    return SimplePosition;
}());
export { SimplePosition };
//# sourceMappingURL=simpleposition.js.map