import { uint32 } from "../core/number";
import { Ordering } from "../core/ordering";
/**
 * Part of a {@link SimplePosition}.
 *
 * A part if a lexicographic triplet (priority, replica, seq).
 * The dot (replica, seq) uniquely identifies.
 */
export declare class SimplePositionPart {
    /**
     * @param priority {@link SimplePositionPart#priority }
     * @param replica {@link SimplePositionPart#replica }
     * @param seq {@link SimplePositionPart#seq }
     */
    protected constructor(priority: uint32, replica: uint32, seq: uint32);
    /**
     * @param priority {@link SimplePositionPart#priority }
     *      ∉ {UINT32_BOTTOM, UINT32_TOP}
     * @param replica {@link SimplePositionPart#replica }
     * @param seq {@link SimplePositionPart#seq }
     */
    static from(priority: uint32, replica: uint32, seq: uint32): SimplePositionPart;
    /**
     * Lowest part.
     * Here the pair (replica, seq) has no meaning.
     */
    static readonly BOTTOM: SimplePositionPart;
    /**
     * Greatest part.
     * Here the pair (replica, seq) has no meaning.
     */
    static readonly TOP: SimplePositionPart;
    /**
     * @param {uint32} seq The seq of the new PositionPart
     * @return {SimplePositionPart} with the same base,
     *  but with a different seq
     */
    withSeq(seq: uint32): SimplePositionPart;
    /**
     * Priority of the part over another one. Lower is better.
     *
     * If two parts have the same priority, the lexicographic pair
     * (replica, seq) is used as disambiguator.
     */
    readonly priority: uint32;
    /**
     * Globally unique identifier of the
     * author which generated this part.
     */
    readonly replica: uint32;
    /**
     * Locally unique sequence number.
     * Each generated part has a sequence number.
     */
    readonly seq: uint32;
    /**
     * @return [priority, replica, seq]
     */
    asTuple(): [uint32, uint32, uint32];
    /**
     * Hash code.
     */
    readonly digest: uint32;
    /**
     * @example
     * SimplePositionPart(0, 0, _) == SimplePositionPart(0, 0, _)
     * SimplePositionPart(0, 0, _) < SimplePositionPart(1, 0, _)
     * SimplePositionPart(1, 0, _) < SimplePositionPart(1, 1, _)
     *
     * Lexicographic order of pair (priority, replica).
     *
     * @param other
     * @return Compare this and {@link other } regardless their sequence
     *  {@link SimplePositionPart#seq}
     */
    compareBase(other: SimplePositionPart): Ordering;
    /**
     * @example
     * SimplePositionPart(0, 0, 0) < SimplePositionPart(1, 0, 0)
     * SimplePositionPart(1, 0, 0) < SimplePositionPart(1, 1, 0)
     * SimplePositionPart(1, 1, 0) < SimplePositionPart(1, 1, 1)
     *
     * Lexicographic order of triple (priority, replica, seq).
     *
     * @param other
     * @return this [Order relation] {@link other}.
     */
    compare(other: SimplePositionPart): Ordering;
}
