"use strict";
/*
    Copyright (C) 2018  Victorien Elvinger

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.
*/
Object.defineProperty(exports, "__esModule", { value: true });
var assert_1 = require("../core/assert");
var number_1 = require("../core/number");
var ordering_1 = require("../core/ordering");
/**
 * Part of a {@link SimplePosition}.
 *
 * A part if a lexicographic triplet (priority, replica, seq).
 * The dot (replica, seq) uniquely identifies.
 */
var SimplePositionPart = /** @class */ (function () {
    /**
     * @param priority {@link SimplePositionPart#priority }
     * @param replica {@link SimplePositionPart#replica }
     * @param seq {@link SimplePositionPart#seq }
     */
    function SimplePositionPart(priority, replica, seq) {
        assert_1.assert(function () { return number_1.isUint32(priority); }, "random ∈ uint32");
        assert_1.assert(function () { return number_1.isUint32(replica); }, "replica ∈ uint32");
        assert_1.assert(function () { return number_1.isUint32(seq); }, "seq ∈ uint32");
        this.priority = priority;
        this.replica = replica;
        this.seq = seq;
    }
    /**
     * @param priority {@link SimplePositionPart#priority }
     *      ∉ {UINT32_BOTTOM, UINT32_TOP}
     * @param replica {@link SimplePositionPart#replica }
     * @param seq {@link SimplePositionPart#seq }
     */
    SimplePositionPart.from = function (priority, replica, seq) {
        assert_1.assert(function () { return priority !== number_1.UINT32_BOTTOM && priority !== number_1.UINT32_TOP; }, "priority ∉ {UINT32_BOTTOM, UINT32_TOP}");
        assert_1.assert(function () { return replica !== number_1.UINT32_TOP; }, "replica != UINT32_TOP. This is reserved for BOTTOM and TOP.");
        return new SimplePositionPart(priority, replica, seq);
    };
    // Derivation
    /**
     * @param {uint32} seq The seq of the new PositionPart
     * @return {SimplePositionPart} with the same base,
     *  but with a different seq
     */
    SimplePositionPart.prototype.withSeq = function (seq) {
        assert_1.assert(function () { return number_1.isUint32(seq); }, "seq ∈ uint32");
        return SimplePositionPart.from(this.priority, this.replica, seq);
    };
    /**
     * @return [priority, replica, seq]
     */
    SimplePositionPart.prototype.asTuple = function () {
        return [this.priority, this.replica, this.seq];
    };
    Object.defineProperty(SimplePositionPart.prototype, "digest", {
        /**
         * Hash code.
         */
        get: function () {
            return number_1.digestOf(this.asTuple());
        },
        enumerable: true,
        configurable: true
    });
    // Status
    /**
     * @example
     * SimplePositionPart(0, 0, _) == SimplePositionPart(0, 0, _)
     * SimplePositionPart(0, 0, _) < SimplePositionPart(1, 0, _)
     * SimplePositionPart(1, 0, _) < SimplePositionPart(1, 1, _)
     *
     * Lexicographic order of pair (priority, replica).
     *
     * @param other
     * @return Compare this and {@link other } regardless their sequence
     *  {@link SimplePositionPart#seq}
     */
    SimplePositionPart.prototype.compareBase = function (other) {
        return ordering_1.lexCompareOrdering(number_1.compareUint32(this.priority, other.priority), number_1.compareUint32(this.replica, other.replica));
    };
    /**
     * @example
     * SimplePositionPart(0, 0, 0) < SimplePositionPart(1, 0, 0)
     * SimplePositionPart(1, 0, 0) < SimplePositionPart(1, 1, 0)
     * SimplePositionPart(1, 1, 0) < SimplePositionPart(1, 1, 1)
     *
     * Lexicographic order of triple (priority, replica, seq).
     *
     * @param other
     * @return this [Order relation] {@link other}.
     */
    SimplePositionPart.prototype.compare = function (other) {
        return (this === other)
            ? 0 /* EQUAL */
            : ordering_1.lexCompareOrdering(this.compareBase(other), number_1.compareUint32(this.seq, other.seq));
    };
    /**
     * Lowest part.
     * Here the pair (replica, seq) has no meaning.
     */
    SimplePositionPart.BOTTOM = new SimplePositionPart(number_1.UINT32_BOTTOM, number_1.UINT32_TOP, 0);
    /**
     * Greatest part.
     * Here the pair (replica, seq) has no meaning.
     */
    SimplePositionPart.TOP = new SimplePositionPart(number_1.UINT32_TOP, number_1.UINT32_TOP, 1);
    return SimplePositionPart;
}());
exports.SimplePositionPart = SimplePositionPart;
//# sourceMappingURL=simplepositionpart.js.map